# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Utilities to handle BIDS inputs."""

from __future__ import annotations

import json
import os
import sys
import typing as ty
import warnings
from pathlib import Path

import pandas as pd

SUPPORTED_AGE_UNITS = (
    'weeks',
    'months',
    'years',
)


def write_bidsignore(deriv_dir):
    # TODO: Port to niworkflows
    bids_ignore = (
        '*.html',
        'logs/',
        'figures/',  # Reports
        '*_xfm.*',  # Unspecified transform files
        '*.surf.gii',  # Unspecified structural outputs
        # Unspecified functional outputs
        '*_boldref.nii.gz',
        '*_bold.func.gii',
        '*_mixing.tsv',
        '*_AROMAnoiseICs.csv',
        '*_timeseries.tsv',
    )
    ignore_file = Path(deriv_dir) / '.bidsignore'

    ignore_file.write_text('\n'.join(bids_ignore) + '\n')


def write_derivative_description(bids_dir, deriv_dir, dataset_links=None):
    from nibabies import __version__

    DOWNLOAD_URL = f'https://github.com/nipreps/nibabies/archive/{__version__}.tar.gz'

    bids_dir = Path(bids_dir)
    deriv_dir = Path(deriv_dir)
    desc = {
        'Name': 'NiBabies: Neuroimaging preprocessing workflows for babies',
        'BIDSVersion': '1.4.0',
        'DatasetType': 'derivative',
        'GeneratedBy': [
            {
                'Name': 'NiBabies',
                'Version': __version__,
                'CodeURL': DOWNLOAD_URL,
            }
        ],
        'HowToAcknowledge': 'TODO',
    }

    # Keys that can only be set by environment
    if 'NIBABIES_DOCKER_TAG' in os.environ:
        desc['GeneratedBy'][0]['Container'] = {
            'Type': 'docker',
            'Tag': f'nipreps/nibabies:{os.environ["NIBABIES_DOCKER_TAG"]}',
        }
    if 'NIBABIES_SINGULARITY_URL' in os.environ:
        desc['GeneratedBy'][0]['Container'] = {
            'Type': 'singularity',
            'URI': os.getenv('NIBABIES_SINGULARITY_URL'),
        }

    # Keys deriving from source dataset
    orig_desc = {}
    fname = bids_dir / 'dataset_description.json'
    if fname.exists():
        orig_desc = json.loads(fname.read_text())

    if 'DatasetDOI' in orig_desc:
        desc['SourceDatasets'] = [
            {'URL': f'https://doi.org/{orig_desc["DatasetDOI"]}', 'DOI': orig_desc['DatasetDOI']}
        ]
    if 'License' in orig_desc:
        desc['License'] = orig_desc['License']
    if dataset_links:
        desc['DatasetLinks'] = {k: str(v) for k, v in dataset_links.items()}

    Path.write_text(deriv_dir / 'dataset_description.json', json.dumps(desc, indent=4))


def extract_entities(file_list):
    """
    Return a dictionary of common entities given a list of files.
    Examples
    --------
    >>> extract_entities('sub-01/anat/sub-01_T1w.nii.gz')
    {'subject': '01', 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    >>> extract_entities(['sub-01/anat/sub-01_T1w.nii.gz'] * 2)
    {'subject': '01', 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    >>> extract_entities(['sub-01/anat/sub-01_run-1_T1w.nii.gz',
    ...                   'sub-01/anat/sub-01_run-2_T1w.nii.gz'])
    {'subject': '01', 'run': [1, 2], 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    """
    from collections import defaultdict

    from bids.layout import parse_file_entities
    from niworkflows.utils.connections import listify

    entities = defaultdict(list)
    for e, v in [
        ev_pair for f in listify(file_list) for ev_pair in parse_file_entities(f).items()
    ]:
        entities[e].append(v)

    def _unique(inlist):
        inlist = sorted(set(inlist))
        if len(inlist) == 1:
            return inlist[0]
        return inlist

    return {k: _unique(v) for k, v in entities.items()}


def validate_input_dir(exec_env, bids_dir, participant_label):
    # Ignore issues and warnings that should not influence NiBabies
    import subprocess
    import tempfile

    validator_config_dict = {
        'ignore': [
            'EVENTS_COLUMN_ONSET',
            'EVENTS_COLUMN_DURATION',
            'TSV_EQUAL_ROWS',
            'TSV_EMPTY_CELL',
            'TSV_IMPROPER_NA',
            'VOLUME_COUNT_MISMATCH',
            'BVAL_MULTIPLE_ROWS',
            'BVEC_NUMBER_ROWS',
            'DWI_MISSING_BVAL',
            'INCONSISTENT_SUBJECTS',
            'INCONSISTENT_PARAMETERS',
            'BVEC_ROW_LENGTH',
            'B_FILE',
            'PARTICIPANT_ID_COLUMN',
            'PARTICIPANT_ID_MISMATCH',
            'TASK_NAME_MUST_DEFINE',
            'PHENOTYPE_SUBJECTS_MISSING',
            'STIMULUS_FILE_MISSING',
            'DWI_MISSING_BVEC',
            'EVENTS_TSV_MISSING',
            'TSV_IMPROPER_NA',
            'ACQTIME_FMT',
            'Participants age 89 or higher',
            'DATASET_DESCRIPTION_JSON_MISSING',
            'FILENAME_COLUMN',
            'WRONG_NEW_LINE',
            'MISSING_TSV_COLUMN_CHANNELS',
            'MISSING_TSV_COLUMN_IEEG_CHANNELS',
            'MISSING_TSV_COLUMN_IEEG_ELECTRODES',
            'UNUSED_STIMULUS',
            'CHANNELS_COLUMN_SFREQ',
            'CHANNELS_COLUMN_LOWCUT',
            'CHANNELS_COLUMN_HIGHCUT',
            'CHANNELS_COLUMN_NOTCH',
            'CUSTOM_COLUMN_WITHOUT_DESCRIPTION',
            'ACQTIME_FMT',
            'SUSPICIOUSLY_LONG_EVENT_DESIGN',
            'SUSPICIOUSLY_SHORT_EVENT_DESIGN',
            'MALFORMED_BVEC',
            'MALFORMED_BVAL',
            'MISSING_TSV_COLUMN_EEG_ELECTRODES',
            'MISSING_SESSION',
        ],
        'error': [],
        'ignoredFiles': ['/dataset_description.json', '/participants.tsv'],
    }
    # Limit validation only to data from requested participants
    if participant_label:
        all_subs = {s.name[4:] for s in bids_dir.glob('sub-*')}
        selected_subs = {s[4:] if s.startswith('sub-') else s for s in participant_label}
        bad_labels = selected_subs.difference(all_subs)
        if bad_labels:
            error_msg = (
                'Data for requested participant(s) label(s) not found. Could '
                'not find data for participant(s): %s. Please verify the requested '
                'participant labels.'
            )
            if exec_env == 'docker':
                error_msg += (
                    ' This error can be caused by the input data not being '
                    'accessible inside the docker container. Please make sure all '
                    'volumes are mounted properly (see https://docs.docker.com/'
                    'engine/reference/commandline/run/#mount-volume--v---read-only)'
                )
            if exec_env == 'singularity':
                error_msg += (
                    ' This error can be caused by the input data not being '
                    'accessible inside the singularity container. Please make sure '
                    'all paths are mapped properly (see https://www.sylabs.io/'
                    'guides/3.0/user-guide/bind_paths_and_mounts.html)'
                )
            raise RuntimeError(error_msg % ','.join(bad_labels))

        ignored_subs = all_subs.difference(selected_subs)
        if ignored_subs:
            for sub in ignored_subs:
                validator_config_dict['ignoredFiles'].append(f'/sub-{sub}/**')
    with tempfile.NamedTemporaryFile(mode='w+', suffix='.json') as temp:
        temp.write(json.dumps(validator_config_dict))
        temp.flush()
        try:
            subprocess.check_call(['bids-validator', str(bids_dir), '-c', temp.name])  # noqa: S607
        except FileNotFoundError:
            print('bids-validator does not appear to be installed', file=sys.stderr)


def parse_bids_for_age_months(
    bids_root: str | Path,
    subject_id: str,
    session_id: str | None = None,
) -> int | None:
    """
    Given a BIDS root, query the BIDS metadata files for participant age, and return in
    chronological months.

    The heuristic followed is:
    1) Check `sub-<subject_id>[/ses-<session_id>]/<sub-<subject_id>[_ses-<session-id>]_scans.tsv
    2) Check `sub-<subject_id>/sub-<subject_id>_sessions.tsv`
    3) Check `<root>/participants.tsv`
    """
    if subject_id.startswith('sub-'):
        subject_id = subject_id[4:]
    if session_id and session_id.startswith('ses-'):
        session_id = session_id[4:]

    # Play nice with sessions
    subject = f'sub-{subject_id}'
    session = f'ses-{session_id}' if session_id else ''
    prefix = f'{subject}' + (f'_{session}' if session else '')

    subject_level = session_level = Path(bids_root) / subject
    if session_id:
        session_level = subject_level / session

    age = None

    scans_tsv = session_level / f'{prefix}_scans.tsv'
    if scans_tsv.exists():
        age = _get_age_from_tsv(
            scans_tsv,
            index_column='filename',
            index_value=r'^anat.*',
        )

    if age is not None:
        return age

    sessions_tsv = subject_level / f'{subject}_sessions.tsv'
    if sessions_tsv.exists() and session_id is not None:
        age = _get_age_from_tsv(sessions_tsv, index_column='session_id', index_value=session)

    if age is not None:
        return age

    participants_tsv = Path(bids_root) / 'participants.tsv'
    if participants_tsv.exists() and age is None:
        age = _get_age_from_tsv(
            participants_tsv, index_column='participant_id', index_value=subject
        )

    return age


def _get_age_from_tsv(
    bids_tsv: Path,
    index_column: str | None = None,
    index_value: str | None = None,
) -> float | None:
    df = pd.read_csv(str(bids_tsv), sep='\t')
    age_col = None

    for column in ('age_weeks', 'age_months', 'age_years', 'age'):
        if column in df.columns:
            age_col = column
            break
    if age_col is None:
        return

    df = df[df[index_column].str.fullmatch(index_value)]

    # Multiple indices may be present after matching
    if len(df) > 1:
        warnings.warn(
            f'Multiple matches for {index_column}:{index_value} found in {bids_tsv.name}.',
            stacklevel=1,
        )

    try:
        # extract age value from row
        age = float(df.loc[df.index[0], age_col].item())
    except Exception:  # noqa: BLE001
        return

    if age_col == 'age':
        # verify age is in months
        bids_json = bids_tsv.with_suffix('.json')
        age_units = _get_age_units(bids_json)
        if age_units is False:
            raise FileNotFoundError(
                f'Could not verify age unit for {bids_tsv.name} - ensure a sidecar JSON '
                'describing column `age` units is available.'
            )
    else:
        age_units = age_col.split('_')[-1]

    age_months = age_to_months(age, units=age_units)
    return age_months


def _get_age_units(bids_json: Path) -> ty.Literal['weeks', 'months', 'years', False]:
    try:
        data = json.loads(bids_json.read_text())
    except (json.JSONDecodeError, OSError):
        return False

    units = data.get('age', {}).get('Units', '')
    if not isinstance(units, str):
        # Multiple units consfuse us
        return False

    if units.lower() in SUPPORTED_AGE_UNITS:
        return units.lower()
    return False


def age_to_months(age: int | float, units: ty.Literal['weeks', 'months', 'years']) -> int:
    """
    Convert a given age, in either "weeks", "months", or "years", into months.

    >>> age_to_months(1, 'years')
    12
    >>> age_to_months(0.5, 'years')
    6
    >>> age_to_months(2, 'weeks')
    0
    >>> age_to_months(3, 'weeks')
    1
    >>> age_to_months(8, 'months')
    8
    """
    WEEKS_TO_MONTH = 0.230137
    YEARS_TO_MONTH = 12

    if units == 'weeks':
        age *= WEEKS_TO_MONTH
    elif units == 'years':
        age *= YEARS_TO_MONTH
    return int(round(age))



#!/usr/bin/env python
"""
This is a Python wrapper to facilitate running NiBabies through Docker or Singularity services.
Docker or Singularity must be installed and running.
To test installation, you can use one of the following:

```bash
docker info
singularity version
```

Please report any feedback to our GitHub repository
(https://github.com/nipreps/nibabies) and do not
forget to credit all the authors of service that NiBabies
uses (https://fmriprep.readthedocs.io/en/latest/citing.html).
"""
import os
import re
import subprocess
import sys

try:
    from ._version import __version__
except ImportError:
    __version__ = '0+unknown'
__copyright__ = 'Copyright 2023, The NiPreps Developers'
__bugreports__ = 'https://github.com/nipreps/nibabies/issues'

MISSING = """
Image '{}' is missing
Would you like to download? [Y/n] """
PKG_PATH = '/opt/conda/envs/nibabies/lib/python3.12/site-packages'
TF_TEMPLATES = (
    'MNI152Lin',
    'MNI152NLin2009cAsym',
    'MNI152NLin6Asym',
    'MNI152NLin6Sym',
    'MNIInfant',
    'MNIPediatricAsym',
    'NKI',
    'OASIS30ANTs',
    'PNC',
    'UNCInfant',
    'fsLR',
    'fsaverage',
    'fsaverage5',
    'fsaverage6',
)
NONSTANDARD_REFERENCES = ('anat', 'T1w', 'run', 'func', 'sbref', 'fsnative')

# Monkey-patch Py2 subprocess
if not hasattr(subprocess, 'DEVNULL'):
    subprocess.DEVNULL = -3

if not hasattr(subprocess, 'run'):
    # Reimplement minimal functionality for usage in this file
    def _run(args, stdout=None, stderr=None):
        from collections import namedtuple

        result = namedtuple('CompletedProcess', 'stdout stderr returncode')

        devnull = None
        if subprocess.DEVNULL in (stdout, stderr):
            devnull = open(os.devnull, 'r+')
            if stdout == subprocess.DEVNULL:
                stdout = devnull
            if stderr == subprocess.DEVNULL:
                stderr = devnull

        proc = subprocess.Popen(args, stdout=stdout, stderr=stderr)
        stdout, stderr = proc.communicate()
        res = result(stdout, stderr, proc.returncode)

        if devnull is not None:
            devnull.close()

        return res

    subprocess.run = _run


# De-fang Python 2's input - we don't eval user input
try:
    input = raw_input
except NameError:
    pass


# The helper class to facilate Docker / Singularity nuiances
class ContainerManager:
    def __init__(self, service, image=None):
        """
        Inputs
        ------
        service : str
            container service to use; either docker or singularity
        """
        self.service = service
        self.image = image
        self.command = [service, 'run']
        self.mounts = []
        if service == 'docker':
            self.add_cmd('--rm')
        elif service == 'singularity':
            self.add_cmd('--cleanenv')

    def add_cmd(self, cmd):
        """Add single or multiple commands to final container run call"""
        if isinstance(cmd, str):
            self.command.append(cmd)
        elif isinstance(cmd, (list, tuple)):
            self.command += list(cmd)

    def add_mount(self, src, dst, read_only=True):
        """
        Generate bind mount string and add to ``mounts`` attribute.

        Inputs
        ------
        src : absolute local path
        dst : absolute container path
        read_only : disable writing to bound path
        """
        self.mounts.append('{0}:{1}{2}'.format(src, dst, ':ro' if read_only else ''))

    def check_install(self):
        """Verify that the service is installed and the user has permission to
        run images.

        Returns
        -------
        -1  Docker/Singularity can't be found
        0  Docker found, but user can't connect to daemon
        1  Test run OK
        """
        try:
            ret = subprocess.run(
                [self.service, 'version'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except OSError as e:
            from errno import ENOENT

            if e.errno == ENOENT:
                return -1
            raise e
        if ret.stderr.startswith(b'Cannot connect to the Docker daemon.'):
            return 0
        return 1

    def check_image(self, image):
        """Check whether image is present on local system"""
        if self.service == 'docker':
            ret = subprocess.run(['docker', 'images', '-q', image], stdout=subprocess.PIPE)
            return bool(ret.stdout)
        elif self.service == 'singularity':
            # check if the image file exists
            return os.path.exists(os.path.abspath(image))
        raise NotImplementedError

    def check_memory(self, image):
        """Check total memory from within an image"""
        if self.service == 'docker':
            cmd = ['docker', 'run', '--rm', '--entrypoint=free', image, '-m']
        elif self.service == 'singularity':
            cmd = ['singularity', 'exec', image, 'free', '-m']
        ret = subprocess.run(cmd, stdout=subprocess.PIPE)
        if ret.returncode:
            return -1
        mem = [
            line.decode().split()[1]
            for line in ret.stdout.splitlines()
            if line.startswith(b'Mem:')
        ][0]
        return int(mem)

    def set_version(self):
        if self.service == 'docker':
            ret = subprocess.run(
                ['docker', 'version', '--format', '{{.Server.Version}}'], stdout=subprocess.PIPE
            )
        elif self.service == 'singularity':
            ret = subprocess.run(['singularity', 'version'], stdout=subprocess.PIPE)
        version = ret.stdout.decode('ascii').strip()
        version_env = '%s_VERSION_8395080871' % self.service.upper()
        self.add_envvar((version_env, version))

    def add_envvar(self, envtuple):
        """Set an environment variable

        Inputs
        ------
        envtuple : tuple in the form of ("ENV_VAR", "value")
        """
        if self.service == 'docker':
            env = '='.join(envtuple)
            self.add_cmd(['-e', env])
        elif self.service == 'singularity':
            # singularity will transfer over environment variables
            # with the prefix: SINGULARITYENV_
            envvar, value = envtuple
            envvar = 'SINGULARITYENV_' + envvar
            os.environ[envvar] = value

    def finalize_container_cmd(self):
        """Add bindings to final command, and finish with image"""
        if self.service == 'docker':
            mflag = '-v'
        elif self.service == 'singularity':
            mflag = '-B'

        for mount in self.mounts:
            self.add_cmd((mflag, mount))
        self.add_cmd(self.image)


def merge_help(wrapper_help, target_help):
    def _get_posargs(usage):
        """
        Extract positional arguments from usage string.

        This function can be used by both native fmriprep (`fmriprep -h`)
        and the docker wrapper (`fmriprep-docker -h`).
        """
        posargs = []
        for targ in usage.split('\n')[-3:]:
            line = targ.lstrip()
            if line.startswith('usage'):
                continue
            if line[0].isalnum() or line[0] == '{':
                posargs.append(line)
            elif line[0] == '[' and (line[1].isalnum() or line[1] == '{'):
                posargs.append(line)
        return ' '.join(posargs)

    # Matches all flags with up to one nested square bracket
    opt_re = re.compile(r'(\[--?[\w-]+(?:[^\[\]]+(?:\[[^\[\]]+\])?)?\])')
    # Matches flag name only
    flag_re = re.compile(r'\[--?([\w-]+)[ \]]')

    # Normalize to Unix-style line breaks
    w_help = wrapper_help.rstrip().replace('\r', '')
    t_help = target_help.rstrip().replace('\r', '')

    w_usage, w_details = w_help.split('\n\n', 1)
    w_groups = w_details.split('\n\n')
    t_usage, t_details = t_help.split('\n\n', 1)
    t_groups = t_details.split('\n\n')

    w_posargs = _get_posargs(w_usage)
    t_posargs = _get_posargs(t_usage)

    w_options = opt_re.findall(w_usage)
    w_flags = sum(map(flag_re.findall, w_options), [])
    t_options = opt_re.findall(t_usage)
    t_flags = sum(map(flag_re.findall, t_options), [])

    # The following code makes this assumption
    # assert w_flags[:2] == ["h", "version"]
    # assert w_posargs.replace("]", "").replace("[", "") == t_posargs

    # Make sure we're not clobbering options we don't mean to
    overlap = set(w_flags).intersection(t_flags)
    expected_overlap = {
        'bids-database-dir',
        'bids-filter-file',
        'derivatives',
        'deriv-filter-file',
        'fs-license-file',
        'fs-subjects-dir',
        'config-file',
        'segmentation-atlases-dir',
        'h',
        'use-plugin',
        'version',
        'w',
    }

    assert overlap == expected_overlap, 'Clobbering options: {}'.format(
        ', '.join(overlap - expected_overlap)
    )

    sections = []

    # Construct usage
    start = w_usage[: w_usage.index(' [')]
    indent = ' ' * len(start)
    new_options = sum(
        (
            w_options[:2],
            [opt for opt, flag in zip(t_options, t_flags, strict=False) if flag not in overlap],
            w_options[2:],
        ),
        [],
    )
    opt_line_length = 79 - len(start)
    length = 0
    opt_lines = [start]
    for opt in new_options:
        opt = ' ' + opt
        olen = len(opt)
        if length + olen <= opt_line_length:
            opt_lines[-1] += opt
            length += olen
        else:
            opt_lines.append(indent + opt)
            length = olen
    opt_lines.append(indent + ' ' + t_posargs)
    sections.append('\n'.join(opt_lines))

    # Use target description and positional args
    sections.extend(t_groups[:2])

    for line in t_groups[2].split('\n')[1:]:
        content = line.lstrip().split(',', 1)[0]
        if content[1:] not in overlap:
            w_groups[2] += '\n' + line

    sections.append(w_groups[2])

    # All remaining sections, show target then wrapper (skipping duplicates)
    sections.extend(t_groups[3:] + w_groups[6:])
    return '\n\n'.join(sections)


def is_in_directory(filepath, directory):
    return os.path.realpath(filepath).startswith(os.path.realpath(directory) + os.sep)


def get_parser():
    """Defines the command line interface of the wrapper"""
    import argparse
    from functools import partial

    class ToDict(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            d = {}
            for kv in values:
                k, v = kv.split('=')
                d[k] = os.path.abspath(v)
            setattr(namespace, self.dest, d)

    def _is_file(path, parser):
        """Ensure a given path exists and it is a file."""
        path = os.path.abspath(path)
        if not os.path.isfile(path):
            raise parser.error('Path should point to a file (or symlink of file): <%s>.' % path)
        return path

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False,
    )

    IsFile = partial(_is_file, parser=parser)

    # require users to specify container service
    parser.add_argument('service', nargs='?', choices=('docker', 'singularity'))
    # Standard NiBabies arguments
    parser.add_argument('bids_dir', nargs='?', type=os.path.abspath, default='')
    parser.add_argument('output_dir', nargs='?', type=os.path.abspath, default='')
    parser.add_argument(
        'analysis_level', nargs='?', choices=['participant'], default='participant'
    )

    parser.add_argument(
        '-h', '--help', action='store_true', help='show this help message and exit'
    )
    parser.add_argument(
        '--version', action='store_true', help="show program's version number and exit"
    )

    # Allow alternative images (semi-developer)
    parser.add_argument(
        '-i',
        '--image',
        metavar='IMG',
        type=str,
        default='nipreps/nibabies:%s' % __version__,
        help='image name',
    )

    # Options for mapping files and directories into container
    # Update `expected_overlap` variable in merge_help() when adding to this
    g_wrap = parser.add_argument_group(
        'Wrapper options',
        'Standard options that require mapping files into the container',
    )
    g_wrap.add_argument(
        '-w',
        '--work-dir',
        action='store',
        type=os.path.abspath,
        help='path where intermediate results should be stored',
    )
    g_wrap.add_argument(
        '--output-spaces',
        nargs='*',
        help="""\
Standard and non-standard spaces to resample anatomical and functional images to. \
Standard spaces may be specified by the form \
``<TEMPLATE>[:res-<resolution>][:cohort-<label>][...]``, where ``<TEMPLATE>`` is \
a keyword (valid keywords: %s) or path pointing to a user-supplied template, and \
may be followed by optional, colon-separated parameters. \
Non-standard spaces (valid keywords: %s) imply specific orientations and sampling \
grids. \
Important to note, the ``res-*`` modifier does not define the resolution used for \
the spatial normalization."""
        % (
            ', '.join('"%s"' % s for s in TF_TEMPLATES),
            ', '.join(NONSTANDARD_REFERENCES),
        ),
    )

    g_wrap.add_argument(
        '--fs-license-file',
        metavar='PATH',
        type=IsFile,
        default=os.getenv('FS_LICENSE', None),
        help='Path to FreeSurfer license key file. Get it (for free) by registering'
        ' at https://surfer.nmr.mgh.harvard.edu/registration.html',
    )
    g_wrap.add_argument(
        '--fs-subjects-dir',
        metavar='PATH',
        type=os.path.abspath,
        help='Path to existing Infant FreeSurfer subjects directory to reuse. ',
    )
    g_wrap.add_argument(
        '--config-file',
        metavar='PATH',
        type=os.path.abspath,
        help='Use pre-generated configuration file. Values in file will be overridden '
        'by command-line arguments.',
    )
    g_wrap.add_argument(
        '--use-plugin',
        metavar='PATH',
        action='store',
        default=None,
        type=os.path.abspath,
        help='nipype plugin configuration file',
    )
    g_wrap.add_argument(
        '--bids-database-dir',
        metavar='PATH',
        type=os.path.abspath,
        help='Path to an existing PyBIDS database folder, for faster indexing '
        '(especially useful for large datasets).',
    )
    g_wrap.add_argument(
        '--segmentation-atlases-dir',
        metavar='PATH',
        type=os.path.abspath,
        help='Directory containing prelabeled segmentations to use for JointLabelFusion.',
    )
    g_wrap.add_argument(
        '-d',
        '--derivatives',
        nargs='+',
        metavar='PATH',
        action=ToDict,
        help='Search PATH(s) for pre-computed derivatives.',
    )
    g_wrap.add_argument(
        '--bids-filter-file',
        metavar='PATH',
        type=os.path.abspath,
        help='Filter file',
    )
    g_wrap.add_argument(
        '--deriv-filter-file',
        metavar='PATH',
        type=os.path.abspath,
        help='Filter file',
    )
    g_wrap.add_argument(
        '--pooch-cache-dir',
        metavar='DIR',
        type=os.path.abspath,
        help='Directory to serve as cache for pooch files'
    )

    # Developer patch/shell options
    g_dev = parser.add_argument_group(
        'Developer options', 'Tools for testing and debugging nibabies'
    )
    g_dev.add_argument(
        '--patch',
        nargs='+',
        metavar='PACKAGE=PATH',
        action=ToDict,
        help='local repository to use within container',
    )
    g_dev.add_argument(
        '--shell',
        action='store_true',
        help='open shell in image instead of running nibabies',
    )
    g_dev.add_argument(
        '--config',
        metavar='PATH',
        action='store',
        type=os.path.abspath,
        help='Use custom nipype.cfg file',
    )
    g_dev.add_argument(
        '-e',
        '--env',
        action='append',
        nargs=2,
        metavar=('ENV_VAR', 'value'),
        help='Set custom environment variable within container',
    )
    g_dev.add_argument(
        '-u',
        '--user',
        action='store',
        help='Run container as a given user/uid. Additionally, group/gid can be'
        'assigned, (i.e., --user <UID>:<GID>)',
    )
    g_dev.add_argument(
        '--network',
        action='store',
        help='Run container with a different network driver '
        '("none" to simulate no internet connection)',
    )
    g_dev.add_argument('--no-tty', action='store_true', help='Run docker without TTY flag -it')

    return parser


def main():
    """Entry point"""

    parser = get_parser()
    # Capture additional arguments to pass inside container
    opts, unknown_args = parser.parse_known_args()

    if opts.version:
        print('nibabies wrapper {!s}'.format(__version__))
        return

    # Set help if no directories set
    if opts.help or not all((opts.service, opts.bids_dir, opts.output_dir)):
        parser.print_help()
        return

    container = ContainerManager(opts.service, image=opts.image)
    check = container.check_install()
    if check < 1:
        if check == -1:
            print(
                'nibabies: Could not find %s command... Is it installed?' % opts.service,
            )
        else:
            print(
                "nibabies: Make sure you have permission to run '%s'" % opts.service,
            )
        return 1

    if not container.check_image(opts.image):
        resp = 'Y'
        if opts.service == 'singularity':
            print('Singularity image must already exist locally.')
            return 1
        try:
            resp = input(MISSING.format(opts.image))
        except KeyboardInterrupt:
            print()
            return 1
        if resp not in ('y', 'Y', ''):
            return 0
        print('Downloading. This may take a while...')

    # Warn on low memory allocation
    mem_total = container.check_memory(opts.image)
    if mem_total == -1:
        print(
            'Could not detect memory capacity of Docker container.\n'
            'Do you have permission to run docker?'
        )
        return 1
    if '--reports-only' not in unknown_args and mem_total < 8000:
        print(
            'Warning: <8GB of RAM is available within your environment.\n'
            'Some parts of nibabies may fail to complete.'
        )
        if '--mem_mb' not in unknown_args:
            resp = 'N'
            try:
                resp = input('Continue anyway? [y/N]')
            except KeyboardInterrupt:
                print()
                return 1
            if resp not in ('y', 'Y', ''):
                return 0

    container.set_version()

    if opts.service == 'docker':
        if not opts.no_tty:
            container.add_cmd('-it')
        if opts.user:
            container.add_cmd(('-u', opts.user))
        if opts.network:
            container.add_cmd('--network=%s' % opts.network)

    # Patch working repositories into installed package directories
    if opts.patch:
        for pkg, repo_path in opts.patch.items():
            container.add_mount(repo_path, os.sep.join((PKG_PATH, pkg)))

    if opts.env:
        for envvar in opts.env:
            container.add_envvar(tuple(envvar))

    if opts.fs_license_file:
        container.add_mount(opts.fs_license_file, '/opt/freesurfer/license.txt')

    main_args = []
    if opts.bids_dir:
        container.add_mount(opts.bids_dir, '/data')
        main_args.append('/data')
    if opts.output_dir:
        if not os.path.exists(opts.output_dir):
            # create it before the container does
            os.makedirs(opts.output_dir)
        container.add_mount(opts.output_dir, '/out', read_only=False)
        main_args.append('/out')
    main_args.append(opts.analysis_level)

    if opts.fs_subjects_dir:
        container.add_mount(opts.fs_subjects_dir, '/opt/subjects', read_only=False)
        unknown_args.extend(['--fs-subjects-dir', '/opt/subjects'])

    if opts.config_file:
        container.add_mount(opts.config_file, '/tmp/config.toml', read_only=False)
        unknown_args.extend(['--config-file', '/tmp/config.toml'])

    if opts.segmentation_atlases_dir:
        container.add_mount(opts.segmentation_atlases_dir, '/opt/segmentations')
        unknown_args.extend(['--segmentation-atlases-dir', '/opt/segmentations'])
    if opts.bids_filter_file:
        container.add_mount(opts.bids_filter_file, '/opt/bids_filters.json')
        unknown_args.extend(['--bids-filter-file', '/opt/bids_filters.json'])
    if opts.deriv_filter_file:
        container.add_mount(opts.deriv_filter_file, '/opt/derivative_filters.json')
        unknown_args.extend(['--deriv-filter-file', '/opt/derivative_filters.json'])
    if opts.pooch_cache_dir:
        container.add_mount(opts.pooch_cache_dir, '/tmp/pooch_cache', read_only=False)
        container.add_envvar(('NIBABIES_POOCH_DIR', '/tmp/pooch_cache'))
    # Patch derivatives for searching
    if opts.derivatives:
        deriv_args = ['--derivatives']
        for deriv, deriv_path in opts.derivatives.items():
            deriv_target = '/deriv/%s' % deriv
            container.add_mount(deriv_path, deriv_target)
            deriv_args.append('='.join([deriv, deriv_target]))
        unknown_args.extend(deriv_args)

    # Check that work_dir is not a child of bids_dir
    if opts.work_dir and opts.bids_dir:
        if is_in_directory(opts.work_dir, opts.bids_dir):
            print(
                'The selected working directory is a subdirectory of the input BIDS folder. '
                'Please modify the output path.'
            )
            return 1

        if not os.path.exists(opts.work_dir):
            # create it before the container does
            os.makedirs(opts.work_dir)
        container.add_mount(opts.work_dir, '/scratch', read_only=False)
        unknown_args.extend(['-w', '/scratch'])

    if opts.config:
        container.add_mount('opts.config', '/home/fmriprep/.nipype/nipype.cfg')

    if opts.use_plugin:
        container.add_mount(opts.use_plugin, '/tmp/plugin.yml')
        unknown_args.extend(['--use-plugin', '/tmp/plugin.yml'])

    if opts.bids_database_dir:
        container.add_mount(opts.bids_database_dir, '/tmp/bids_db', read_only=False)
        unknown_args.extend(['--bids-database-dir', '/tmp/bids_db'])

    if opts.output_spaces:
        spaces = []
        for space in opts.output_spaces:
            if space.split(':')[0] not in (TF_TEMPLATES + NONSTANDARD_REFERENCES):
                tpl = os.path.basename(space)
                if not tpl.startswith('tpl-'):
                    raise RuntimeError('Custom template %s requires a `tpl-` prefix' % tpl)
                target = '/home/fmriprep/.cache/templateflow/' + tpl
                container.add_mount(os.path.abspath(space), target)
                spaces.append(tpl[4:])
            else:
                spaces.append(space)
        unknown_args.extend(['--output-spaces'] + spaces)

    if opts.shell:
        if opts.service == 'docker':
            container.add_cmd('--entrypoint=bash')
        elif opts.service == 'singularity':
            # replace default "run" command
            container.command[1] = 'shell'

    container.image = opts.image
    # after this, all call to ``container.add_cmd``
    # will be for nibabies arguments
    container.finalize_container_cmd()

    # Override help and version to describe underlying program
    # Respects '-i' flag, so will retrieve information from any image
    # if opts.help:
    #     container.add_cmd("-h")
    #     targethelp = subprocess.check_output(container.command).decode()
    #     print(merge_help(parser.format_help(), targethelp))
    #     return 0
    # elif opts.version:
    #     # Get version to be run and exit
    #     container.add_cmd("--version")
    #     ret = subprocess.run(container.command)
    #     return ret.returncode

    if not opts.shell:
        container.add_cmd(main_args)
        container.add_cmd(unknown_args)

    print('RUNNING: ' + ' '.join(container.command))
    ret = subprocess.run(container.command)
    if ret.returncode:
        print('nibabies: Please report errors to %s' % __bugreports__)
    return ret.returncode


if __name__ == '__main__':
    if '__main__.py' in sys.argv[0]:
        from . import __name__ as module

        sys.argv[0] = '%s -m %s' % (sys.executable, module)
    sys.exit(main())



"""Signal processing filters."""


def truncation(
    in_file,
    clip_max=99.9,
    dtype='int16',
    out_file=None,
    out_max=1000,
    out_min=0,
    percentiles=(0.1, 95),
):
    """Truncate and clip the input image intensities."""
    from pathlib import Path

    import nibabel as nb
    import numpy as np
    from nipype.utils.filemanip import fname_presuffix

    try:
        info = np.iinfo(dtype)
    except ValueError:
        info = np.finfo(dtype)

    img = nb.load(in_file)
    hdr = img.header.copy()
    hdr.set_data_dtype(dtype)

    data = img.get_fdata()

    out_min = max(out_min, info.min)
    out_max = min(out_max, info.max)

    a_min = np.percentile(data.reshape(-1), percentiles[0])
    data -= a_min
    a_max = np.percentile(data.reshape(-1), percentiles[1])
    data *= out_max / a_max
    data = np.clip(data, info.min, info.max)

    if clip_max is not None:
        data = np.clip(data, 0, np.percentile(data.reshape(-1), clip_max))

    if out_file is None:
        out_file = fname_presuffix(Path(in_file).name, suffix='_trunc')

    out_file = str(Path(out_file).absolute())
    img.__class__(data.astype(dtype), img.affine, hdr).to_filename(out_file)
    return out_file


def gaussian_filter(in_file, sigma=None, out_file=None):
    """Filter input image by convolving with a Gaussian kernel."""
    from pathlib import Path

    import nibabel as nb
    import numpy as np
    from nipype.utils.filemanip import fname_presuffix
    from scipy.ndimage import gaussian_filter

    if out_file is None:
        out_file = fname_presuffix(Path(in_file).name, suffix='_gauss')
    out_file = str(Path(out_file).absolute())

    img = nb.load(in_file)
    if sigma is None:
        sigma = tuple(np.array(img.header.get_zooms()[:3]) * 2.0)
    img.__class__(gaussian_filter(img.dataobj, sigma), img.affine, img.header).to_filename(
        out_file
    )
    return out_file



from pathlib import Path

import nibabel as nb
import numpy as np
import pytest
from nibabel.orientations import aff2axcodes

from nibabies.workflows.anatomical.preproc import (
    _normalize_roi,
    init_conform_derivative_wf,
    init_csf_norm_wf,
)

EXPECTED_CSF_NORM = np.array([[[49, 75], [23, 75]], [[77, 80], [33, 3]]], dtype='uint8')


@pytest.fixture
def anat_file(tmp_path):
    data = np.array([[[49, 73], [23, 73]], [[77, 80], [33, 3]]], dtype='uint8')
    img = nb.Nifti1Image(data, np.eye(4))
    out = tmp_path / 'input.nii.gz'
    img.to_filename(out)
    return out


def test_csf_norm_wf(tmp_path, anat_file):
    tpms = []
    for tpm, thresh in (('gm', 25), ('wm', 75), ('csf', 50)):
        name = tmp_path / f'{tpm}.nii.gz'
        anat_img = nb.load(anat_file)
        anat_data = np.asanyarray(nb.load(anat_file).dataobj)

        binmask = anat_data > thresh
        masked = (binmask * 1).astype('uint8')
        mask = nb.Nifti1Image(masked, anat_img.affine)
        mask.to_filename(name)
        tpms.append(name)

    wf = init_csf_norm_wf()
    wf.base_dir = tmp_path
    wf.inputs.inputnode.anat_preproc = anat_file
    wf.inputs.inputnode.anat_tpms = tpms

    # verify workflow runs
    wf.run()

    # verify function works as expected
    outfile = _normalize_roi(anat_file, tpms[2])
    assert np.array_equal(
        np.asanyarray(nb.load(outfile).dataobj),
        EXPECTED_CSF_NORM,
    )
    Path(outfile).unlink()


@pytest.mark.parametrize(
    ('affine_mismatch', 'ornt_mismatch'),
    [
        (False, False),
        (True, False),
        (False, True),
        (True, True),
    ],
)
def test_conform_derivative_wf(tmp_path, anat_file, affine_mismatch, ornt_mismatch):
    deriv = tmp_path / 'mask.nii.gz'
    ref_img = nb.load(anat_file)
    aff = ref_img.affine.copy()
    if affine_mismatch:
        # Alter affine slightly
        aff[:3, :3] += 0.01
        assert not np.array_equal(aff, ref_img.affine)

    img = ref_img.__class__(ref_img.dataobj, affine=aff)
    if ornt_mismatch:
        from niworkflows.interfaces.nibabel import reorient_image

        img = reorient_image(img, target_ornt='LPI')
        assert aff2axcodes(img.affine) != aff2axcodes(ref_img.affine)

    img.to_filename(deriv)
    wf = init_conform_derivative_wf(in_file=deriv)
    wf.base_dir = tmp_path
    wf.inputs.inputnode.ref_file = anat_file

    wf.run()

    output = list((tmp_path / 'conform_derivative_wf' / 'match_header').glob('*.nii.gz'))
    assert output
    out_file = output[0]
    out_img = nb.load(out_file)
    assert np.array_equal(out_img.affine, ref_img.affine)
    assert aff2axcodes(out_img.affine) == aff2axcodes(ref_img.affine)



from __future__ import annotations

import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe
from niworkflows.interfaces.nibabel import GenerateSamplingReference
from niworkflows.interfaces.utility import KeySelect

from nibabies.interfaces.resampling import (
    DistortionParameters,
    ReconstructFieldmap,
    ResampleSeries,
)


def init_bold_volumetric_resample_wf(
    *,
    metadata: dict,
    mem_gb: dict[str, float],
    jacobian: bool,
    fieldmap_id: str | None = None,
    omp_nthreads: int = 1,
    name: str = 'bold_volumetric_resample_wf',
) -> pe.Workflow:
    """Resample a BOLD series to a volumetric target space.

    This workflow collates a sequence of transforms to resample a BOLD series
    in a single shot, including motion correction and fieldmap correction, if
    requested.

    .. workflow::

        from fmriprep.workflows.bold.resampling import init_bold_volumetric_resample_wf
        wf = init_bold_volumetric_resample_wf(
            metadata={
                'RepetitionTime': 2.0,
                'PhaseEncodingDirection': 'j-',
                'TotalReadoutTime': 0.03
            },
            fieldmap_id='my_fieldmap',
        )

    Parameters
    ----------
    metadata
        BIDS metadata for BOLD file.
    fieldmap_id
        Fieldmap identifier, if fieldmap correction is to be applied.
    omp_nthreads
        Maximum number of threads an individual process may use.
    name
        Name of workflow (default: ``bold_volumetric_resample_wf``)

    Inputs
    ------
    bold_file
        BOLD series to resample.
    bold_ref_file
        Reference image to which BOLD series is aligned.
    target_ref_file
        Reference image defining the target space.
    target_mask
        Brain mask corresponding to ``target_ref_file``.
        This is used to define the field of view for the resampled BOLD series.
    motion_xfm
        List of affine transforms aligning each volume to ``bold_ref_file``.
        If undefined, no motion correction is performed.
    boldref2fmap_xfm
        Affine transform from ``bold_ref_file`` to the fieldmap reference image.
    fmap_ref
        Fieldmap reference image defining the valid field of view for the fieldmap.
    fmap_coeff
        B-Spline coefficients for the fieldmap.
    fmap_id
        Fieldmap identifier, to select correct fieldmap in case there are multiple.
    boldref2anat_xfm
        Affine transform from ``bold_ref_file`` to the anatomical reference image.
    anat2std_xfm
        Affine transform from the anatomical reference image to standard space.
        Leave undefined to resample to anatomical reference space.

    Outputs
    -------
    bold_file
        The ``bold_file`` input, resampled to ``target_ref_file`` space.
    resampling_reference
        An empty reference image with the correct affine and header for resampling
        further images into the BOLD series' space.

    """
    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold_file',
                'bold_ref_file',
                'target_ref_file',
                'target_mask',
                # HMC
                'motion_xfm',
                # SDC
                'boldref2fmap_xfm',
                'fmap_ref',
                'fmap_coeff',
                'fmap_id',
                # Anatomical
                'boldref2anat_xfm',
                # Template
                'anat2std_xfm',
                # Entity for selecting target resolution
                'resolution',
            ],
        ),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['bold_file', 'resampling_reference']),
        name='outputnode',
    )

    gen_ref = pe.Node(GenerateSamplingReference(), name='gen_ref', mem_gb=0.3)

    boldref2target = pe.Node(niu.Merge(2), name='boldref2target', run_without_submitting=True)
    bold2target = pe.Node(niu.Merge(2), name='bold2target', run_without_submitting=True)
    resample = pe.Node(
        ResampleSeries(jacobian=jacobian, num_threads=omp_nthreads),
        name='resample',
        n_procs=omp_nthreads,
        mem_gb=mem_gb['resampled'],
    )

    workflow.connect([
        (inputnode, gen_ref, [
            ('bold_ref_file', 'moving_image'),
            ('target_ref_file', 'fixed_image'),
            ('target_mask', 'fov_mask'),
            (('resolution', _is_native), 'keep_native'),
        ]),
        (inputnode, boldref2target, [
            ('boldref2anat_xfm', 'in1'),
            ('anat2std_xfm', 'in2'),
        ]),
        (inputnode, bold2target, [('motion_xfm', 'in1')]),
        (inputnode, resample, [('bold_file', 'in_file')]),
        (gen_ref, resample, [('out_file', 'ref_file')]),
        (boldref2target, bold2target, [('out', 'in2')]),
        (bold2target, resample, [('out', 'transforms')]),
        (gen_ref, outputnode, [('out_file', 'resampling_reference')]),
        (resample, outputnode, [('out_file', 'bold_file')]),
    ])  # fmt:skip

    if not fieldmap_id:
        return workflow

    fmap_select = pe.Node(
        KeySelect(fields=['fmap_ref', 'fmap_coeff'], key=fieldmap_id),
        name='fmap_select',
        run_without_submitting=True,
    )
    distortion_params = pe.Node(
        DistortionParameters(metadata=metadata),
        name='distortion_params',
        run_without_submitting=True,
    )
    fmap2target = pe.Node(niu.Merge(2), name='fmap2target', run_without_submitting=True)
    inverses = pe.Node(
        niu.Function(function=_gen_inverses),
        name='inverses',
        run_without_submitting=True,
    )

    fmap_recon = pe.Node(ReconstructFieldmap(), name='fmap_recon', mem_gb=1)

    workflow.connect([
        (inputnode, fmap_select, [
            ('fmap_ref', 'fmap_ref'),
            ('fmap_coeff', 'fmap_coeff'),
            ('fmap_id', 'keys'),
        ]),
        (inputnode, distortion_params, [('bold_file', 'in_file')]),
        (inputnode, fmap2target, [('boldref2fmap_xfm', 'in1')]),
        (gen_ref, fmap_recon, [('out_file', 'target_ref_file')]),
        (boldref2target, fmap2target, [('out', 'in2')]),
        (boldref2target, inverses, [('out', 'inlist')]),
        (fmap_select, fmap_recon, [
            ('fmap_coeff', 'in_coeffs'),
            ('fmap_ref', 'fmap_ref_file'),
        ]),
        (fmap2target, fmap_recon, [('out', 'transforms')]),
        (inverses, fmap_recon, [('out', 'inverse')]),
        # Inject fieldmap correction into resample node
        (distortion_params, resample, [
            ('readout_time', 'ro_time'),
            ('pe_direction', 'pe_dir'),
        ]),
        (fmap_recon, resample, [('out_file', 'fieldmap')]),
    ])  # fmt:skip

    return workflow


def _gen_inverses(inlist: list) -> list[bool]:
    """Create a list indicating the first transform should be inverted.

    The input list is the collection of transforms that follow the
    inverted one.
    """
    from niworkflows.utils.connections import listify

    if not inlist:
        return [True]
    return [True] + [False] * len(listify(inlist))


def _is_native(value):
    return value == 'native'



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Head-Motion Estimation and Correction (HMC) of BOLD images
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: init_bold_hmc_wf

"""

from nipype.interfaces import fsl
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe

from ...config import DEFAULT_MEMORY_MIN_GB


def init_bold_hmc_wf(mem_gb: float, omp_nthreads: int, name: str = 'bold_hmc_wf'):
    """
    Build a workflow to estimate head-motion parameters.

    This workflow estimates the motion parameters to perform
    :abbr:`HMC (head motion correction)` over the input
    :abbr:`BOLD (blood-oxygen-level dependent)` image.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.bold import init_bold_hmc_wf
            wf = init_bold_hmc_wf(
                mem_gb=3,
                omp_nthreads=1)

    Parameters
    ----------
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    omp_nthreads : :obj:`int`
        Maximum number of threads an individual process may use
    name : :obj:`str`
        Name of workflow (default: ``bold_hmc_wf``)

    Inputs
    ------
    bold_file
        BOLD series NIfTI file
    raw_ref_image
        Reference image to which BOLD series is motion corrected

    Outputs
    -------
    xforms
        ITKTransform file aligning each volume to ``ref_image``
    movpar_file
        MCFLIRT motion parameters, normalized to SPM format (X, Y, Z, Rx, Ry, Rz)
    rmsd_file
        Root mean squared deviation as measured by ``fsl_motion_outliers`` [Jenkinson2002]_.

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.confounds import NormalizeMotionParams
    from niworkflows.interfaces.itk import MCFLIRT2ITK

    workflow = Workflow(name=name)
    workflow.__desc__ = """\
Head-motion parameters with respect to the BOLD reference
(transformation matrices, and six corresponding rotation and translation
parameters) are estimated before any spatiotemporal filtering using
`mcflirt` [FSL {fsl_ver}, @mcflirt].
""".format(fsl_ver=fsl.Info().version() or '<ver>')

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['bold_file', 'raw_ref_image']), name='inputnode'
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['xforms', 'movpar_file', 'rmsd_file']), name='outputnode'
    )

    # Head motion correction (hmc)
    mcflirt = pe.Node(
        fsl.MCFLIRT(save_mats=True, save_plots=True, save_rms=True),
        name='mcflirt',
        mem_gb=mem_gb * 3,
    )

    fsl2itk = pe.Node(MCFLIRT2ITK(), name='fsl2itk', mem_gb=0.05, n_procs=omp_nthreads)

    normalize_motion = pe.Node(
        NormalizeMotionParams(format='FSL'), name='normalize_motion', mem_gb=DEFAULT_MEMORY_MIN_GB
    )

    def _pick_rel(rms_files):
        return rms_files[-1]

    workflow.connect([
        (inputnode, mcflirt, [('raw_ref_image', 'ref_file'),
                              ('bold_file', 'in_file')]),
        (inputnode, fsl2itk, [('raw_ref_image', 'in_source'),
                              ('raw_ref_image', 'in_reference')]),
        (mcflirt, fsl2itk, [('mat_file', 'in_files')]),
        (mcflirt, normalize_motion, [('par_file', 'in_file')]),
        (mcflirt, outputnode, [(('rms_files', _pick_rel), 'rmsd_file')]),
        (fsl2itk, outputnode, [('out_file', 'xforms')]),
        (normalize_motion, outputnode, [('out_file', 'movpar_file')]),
    ])  # fmt:skip

    return workflow



from niworkflows.utils.testing import generate_bids_skeleton

from nibabies.utils.derivatives import (
    collect_anatomical_derivatives,
    collect_functional_derivatives,
    copy_derivatives,
)

from . import DERIV_SKELETON


def test_collect_derivatives(tmp_path):
    deriv_dir = tmp_path / 'derivatives'
    generate_bids_skeleton(deriv_dir, str(DERIV_SKELETON))
    output_spaces = ['MNIInfant:cohort-1']

    anat_cache = collect_anatomical_derivatives(
        derivatives_dir=deriv_dir,
        subject_id='01',
        session_id=None,
        std_spaces=output_spaces,
    )
    for suffix in ('preproc', 'mask', 'dseg'):
        assert anat_cache[f't2w_{suffix}']
    assert len(anat_cache['t2w_tpms']) == 3
    xfms = anat_cache['transforms']
    for space in output_spaces:
        assert xfms[space]['reverse']
        assert xfms[space]['forward']
    for surface in (
        'white',
        'pial',
        'midthickness',
        'sphere',
        'thickness',
        'sulc',
        'sphere_reg',
        'sphere_reg_fsLR',
    ):
        assert len(anat_cache[surface]) == 2

    func_cache = collect_functional_derivatives(deriv_dir, {'subject': '01'}, None)
    for val in ('hmc_boldref', 'coreg_boldref', 'hmc'):
        assert func_cache[val]


def test_copy_derivatives(tmp_path):
    precomp = tmp_path / 'precomputed'
    precomp.mkdir()
    out = tmp_path / 'out'
    out.mkdir()

    mask = precomp / 'mask.nii.gz'
    mask.touch()
    aseg = precomp / 'aseg.nii.gz'
    aseg.touch()
    aseg_meta = precomp / 'aseg.json'
    aseg_meta.touch()

    derivs = {
        't2w_mask': str(mask),
        't2w_aseg': str(aseg),
        'transforms': {},
    }

    copy_derivatives(derivs, out, 'anat', 'sub-01')
    outpath = out / 'sub-01' / 'anat'

    for fl in (mask, aseg, aseg_meta):
        assert (outpath / fl.name).exists()

    copy_derivatives(derivs, out, 'anat', 'sub-01', 'ses-a')
    outpath = out / 'sub-01' / 'ses-a' / 'anat'

    for fl in (mask, aseg, aseg_meta):
        assert (outpath / fl.name).exists()
